# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the Gaussian process-based sampling."""
from enum import Enum
from typing import Optional, Tuple, cast

import GPy
import numpy as np
from GPy.models import GPRegression
from numpy.typing import NDArray
from scipy.special import erfc  # pylint: disable=no-name-in-module

from black_it.samplers.base import BaseSampler
from black_it.samplers.random_uniform import RandomUniformSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import digitize_data


class _AcquisitionTypes(Enum):
    """Enumeration of allowed acquisition types."""

    MEAN = "mean"
    EI = "expected_improvement"

    def __str__(self) -> str:
        """Get the string representation."""
        return self.value


class GaussianProcessSampler(BaseSampler):
    """
    This class implements the Gaussian process-based sampler.

    In particular, the sampling is based on a Gaussian Process interpolation of the loss function.

    Note: this class is a wrapper of the GPRegression model of the GPy package.
    """

    def __init__(
        self,
        batch_size: int,
        internal_seed: int = 0,
        candidate_pool_size: Optional[int] = None,
        max_iters: int = 1000,
        optimize_restarts: int = 5,
        acquisition: str = "expected_improvement",
    ):
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            internal_seed: the internal state of the sampler, fixing this numbers the sampler behaves deterministically
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            max_iters: maximum number of iteration in the optimization of the GP hyperparameters
            optimize_restarts: number of independent random trials of the optimization of the GP hyperparameters
            acquisition: type of acquisition function, it can be 'expected_improvement' of simply 'mean'
        """
        self._validate_acquisition(acquisition)

        super().__init__(batch_size, internal_seed)
        self.max_iters = max_iters
        self.optimize_restarts = optimize_restarts
        self.acquisition = acquisition
        self._gpmodel: Optional[GPRegression] = None

        self.candidate_pool_size = (
            candidate_pool_size
            if candidate_pool_size is not None
            else 1000 * batch_size
        )

    @staticmethod
    def _validate_acquisition(acquisition: str) -> None:
        """
        Check that the required acquisition is among the supported ones.

        Args:
            acquisition: the acquisition provided as input of the constructor.

        Raises
            ValueError: if the provided acquisition type is not among the allowed ones.
        """
        try:
            _AcquisitionTypes(acquisition)
        except ValueError as e:
            raise ValueError(
                "expected one of the following acquisition types: "
                f"[{' '.join(map(str, _AcquisitionTypes))}], "
                f"got {acquisition}"
            ) from e

    def single_sample(  # noqa
        self,
        seed: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        raise NotImplementedError(
            "single_sample is not supported by GaussianProcessSampler"
        )

    def sample(
        self,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample from the search space.

        Args:
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters
        """
        np.random.seed(self.internal_seed)

        X, Y = existing_points, np.atleast_2d(existing_losses).T

        if X.shape[0] > 500:
            raise RuntimeWarning(
                "Standard GP evaluations can be expensive for large datasets, consider implementing a sparse GP"
            )

        # initialize GP class from GPy with a Matern kernel by default
        kern = GPy.kern.Matern52(search_space.dims, variance=1.0, ARD=False)
        noise_var = Y.var() * 0.01

        self._gpmodel = GPRegression(
            X, Y, kernel=kern, noise_var=noise_var, mean_function=None
        )

        # Make sure we do not get ridiculously small residual noise variance
        self._gpmodel.Gaussian_noise.constrain_bounded(
            1e-9, 1e6, warning=False
        )  # constrain_positive(warning=False)

        if self.max_iters > 0:
            # --- update the model maximizing the marginal likelihood.
            if self.optimize_restarts == 1:
                self._gpmodel.optimize(
                    optimizer="bfgs",
                    max_iters=self.max_iters,
                    messages=False,
                    ipython_notebook=False,
                )
            else:
                self._gpmodel.optimize_restarts(
                    num_restarts=self.optimize_restarts,
                    optimizer="bfgs",
                    max_iters=self.max_iters,
                    verbose=False,
                )

        # Get large candidate pool
        candidates: NDArray[np.float64] = RandomUniformSampler(
            batch_size=self.candidate_pool_size, internal_seed=self.internal_seed
        ).sample(search_space, existing_points, existing_losses)

        # predict mean or expected improvement on the full sample set
        if self.acquisition == _AcquisitionTypes.EI.value:
            # minus sign needed for subsequent sorting
            candidates_score = -self._predict_EI(candidates)[:, 0]
        else:  # acquisition is "mean"
            candidates_score = self._predict_mean_std(candidates)[0][:, 0]

        sorting_indices = np.argsort(candidates_score)

        sampled_points = candidates[sorting_indices][: self.batch_size, :]

        self.internal_seed += self.batch_size

        return digitize_data(sampled_points, search_space.param_grid)

    def _predict_mean_std(
        self, X: NDArray[np.float64]
    ) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Predict mean and standard deviation of a fitted GP.

        Args:
            X: the points on which the predictions should be performed

        Returns:
            The pair (mean, std).
        """
        gpmodel = cast(GPRegression, self._gpmodel)
        X = X[None, :] if X.ndim == 1 else X
        m, v = gpmodel.predict(X, full_cov=False, include_likelihood=True)
        v = np.clip(v, 1e-10, np.inf)
        return m, np.sqrt(v)

    def _get_fmin(self) -> float:
        """Return the location where the posterior mean is takes its minimal value."""
        gpmodel = cast(GPRegression, self._gpmodel)
        return gpmodel.predict(gpmodel.X)[0].min()

    def _predict_EI(
        self, X: NDArray[np.float64], jitter: float = 1e-3
    ) -> NDArray[np.float64]:
        """
        Compute the Expected Improvement per unit of cost.

        Args:
            X:  the points on which the predictions should be performed
            jitter: positive value to make the acquisition more explorative.

        Returns:
            the expected improvement.
        """
        m, s = self._predict_mean_std(X)

        fmin = self._get_fmin()

        phi, Phi, u = get_quantiles(jitter, fmin, m, s)

        f_acqu = s * (u * Phi + phi)

        return f_acqu


def get_quantiles(
    acquisition_par: float, fmin: float, m: NDArray[np.float64], s: NDArray[np.float64]
) -> Tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """
    Quantiles of the Gaussian distribution useful to determine the acquisition function values.

    Args:
        acquisition_par: parameter of the acquisition function
        fmin: current minimum.
        m: vector of means.
        s: vector of standard deviations.

    Returns:
        the quantiles.
    """
    # remove values of variance that are too small
    s[s < 1e-10] = 1e-10

    u: NDArray[np.float64] = (fmin - m - acquisition_par) / s
    phi: NDArray[np.float64] = np.exp(-0.5 * u**2) / np.sqrt(2 * np.pi)
    Phi: NDArray[np.float64] = 0.5 * erfc(-u / np.sqrt(2))

    return phi, Phi, u



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the random uniform sampler."""

import numpy as np
from numpy import random
from numpy.random import default_rng
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace


class RandomUniformSampler(BaseSampler):
    """Random uniform sampling."""

    def single_sample(
        self,
        seed: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample a single point uniformly within the search space.

        Args:
            seed: random seed
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled (not used)
            existing_losses: the loss corresponding to the sampled parameters (not used)

        Returns:
            the parameter sampled
        """
        random_generator: random.Generator = default_rng(seed)

        sampled_point: NDArray[np.float64] = np.zeros(
            shape=search_space.dims, dtype=np.float64
        )

        for i, params in enumerate(search_space.param_grid):
            sampled_point[i] = random_generator.choice(params)

        return sampled_point



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains serialization and deserialization of calibration state with Pandas."""
import gzip
import json
import pickle  # nosec B403
from pathlib import Path
from typing import Optional, Sequence, Tuple

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from black_it.loss_functions.base import BaseLoss
from black_it.samplers.base import BaseSampler
from black_it.utils.base import NumpyArrayEncoder, PathLike


def load_calibrator_state(checkpoint_path: PathLike, _code_state_version: int) -> Tuple:
    """
    Load calibrator data from a given folder.

    Args:
        checkpoint_path: the folder where the data are stored
        _code_state_version: the serialization version (unused)

    Returns:
        all the data needed to reconstruct the calibrator state
    """
    checkpoint_path = Path(checkpoint_path)
    with open(checkpoint_path / "calibration_params.json", "r") as f:
        cp = json.load(f)

    cr = pd.read_csv(checkpoint_path / "calibration_results.csv")

    params_samp_list = []

    for i in range(len(cp["parameters_precision"])):
        params_samp_list.append(cr[f"params_samp_{i}"])

    params_samp = np.vstack(params_samp_list).T

    with open(checkpoint_path / "samplers_pickled.pickle", "rb") as fb:
        samplers = pickle.load(fb)  # nosec B301

    with open(checkpoint_path / "loss_function_pickled.pickle", "rb") as fb:
        loss_function = pickle.load(fb)  # nosec B301

    with gzip.GzipFile(checkpoint_path / "series_samp.npy.gz", "rb") as gzf:
        series_samp = np.load(gzf)

    return (
        # initialization parameters
        np.asarray(cp["parameters_bounds"]),
        np.asarray(cp["parameters_precision"]),
        np.asarray(cp["real_data"]),
        cp["ensemble_size"],
        cp["N"],
        cp["D"],
        cp["convergence_precision"],
        cp["verbose"],
        cp["saving_file"],
        cp["model_seed"],
        cp["model_name"],
        samplers,
        loss_function,
        cp["current_batch_index"],
        cp["n_sampled_params"],
        cp["n_jobs"],
        # calibration results
        params_samp,
        cr["losses_samp"].to_numpy(),
        series_samp,
        cr["batch_num_samp"].to_numpy(),
        cr["method_samp"].to_numpy(),
    )


def save_calibrator_state(  # pylint: disable=too-many-arguments,too-many-locals
    checkpoint_path: PathLike,
    parameters_bounds: NDArray[np.float64],
    parameters_precision: NDArray[np.float64],
    real_data: NDArray[np.float64],
    ensemble_size: int,
    N: int,
    D: int,
    convergence_precision: Optional[float],
    verbose: bool,
    saving_file: Optional[str],
    model_seed: int,
    model_name: str,
    samplers: Sequence[BaseSampler],
    loss_function: BaseLoss,
    current_batch_index: int,
    n_sampled_params: int,
    n_jobs: int,
    params_samp: NDArray[np.float64],
    losses_samp: NDArray[np.float64],
    series_samp: NDArray[np.float64],
    batch_num_samp: NDArray[np.int64],
    method_samp: NDArray[np.int64],
) -> None:
    """
    Store the state of the calibrator in a given folder.

    Args:
        checkpoint_path: path to the checkpoint
        parameters_bounds: the parameters bounds
        parameters_precision: the parameters precision
        real_data: the real data
        ensemble_size: the ensemble size
        N: the number of samples
        D: the number of dimensions
        convergence_precision: the convergence precision
        verbose: the verbosity mode
        saving_file: the saving file
        model_seed: the model seed
        model_name: the model name
        samplers: the ordered list of samplers to use in the calibration
        loss_function: the loss function
        current_batch_index: the current batch index
        n_sampled_params: the number of sampled params
        n_jobs: the number of jobs
        params_samp: the sampled parameters
        losses_samp: the sampled losses
        series_samp: the sampled series
        batch_num_samp: the sampling batch number
        method_samp: the sampling method
    """
    checkpoint_path = Path(checkpoint_path)
    # create directory if needed
    if not checkpoint_path.exists():
        checkpoint_path.mkdir(parents=True)

    calibration_params = dict(
        parameters_bounds=parameters_bounds,
        parameters_precision=parameters_precision,
        real_data=real_data,
        ensemble_size=ensemble_size,
        N=N,
        D=D,
        convergence_precision=convergence_precision,
        verbose=verbose,
        saving_file=saving_file,
        model_seed=model_seed,
        model_name=model_name,
        current_batch_index=current_batch_index,
        n_sampled_params=n_sampled_params,
        n_jobs=n_jobs,
    )
    # save calibration parameters in a json dictionary
    with open(checkpoint_path / "calibration_params.json", "w") as f:
        json.dump(calibration_params, f, cls=NumpyArrayEncoder)

    # save instantiated samplers and loss functions
    with open(checkpoint_path / "samplers_pickled.pickle", "wb") as fb:
        pickle.dump(samplers, fb)

    with open(checkpoint_path / "loss_function_pickled.pickle", "wb") as fb:
        pickle.dump(loss_function, fb)

    # save calibration results into pandas dataframe
    calibration_results = dict(
        losses_samp=losses_samp.tolist(),
        batch_num_samp=batch_num_samp.tolist(),
        method_samp=method_samp.tolist(),
    )

    for d in range(params_samp.shape[1]):
        param_key_d = f"params_samp_{d}"
        calibration_results[param_key_d] = params_samp[:, d]

    df_calibration_results = pd.DataFrame.from_dict(calibration_results)
    df_calibration_results.to_csv(checkpoint_path / "calibration_results.csv")

    # all time series sampled (heavy)
    with gzip.GzipFile(checkpoint_path / "series_samp.npy.gz", mode="wb") as gzf:
        np.save(gzf, series_samp)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the random forest sampling."""
from typing import Optional, Tuple

import numpy as np
from numpy.random import default_rng
from numpy.typing import NDArray
from sklearn.ensemble import RandomForestClassifier

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import digitize_data


class RandomForestSampler(BaseSampler):
    """This class implements random forest sampling."""

    def __init__(
        self,
        batch_size: int,
        internal_seed: int = 0,
        candidate_pool_size: Optional[int] = None,
        n_estimators: int = 500,
        criterion: str = "gini",
    ) -> None:
        """
        Random forest sampling.

        Note: this class is a wrapper of sklearn.ensemble.RandomForestClassifier.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            internal_seed: the internal state of the sampler, fixing this numbers the sampler behaves deterministically
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            n_estimators: number of trees in the forest
            criterion: The function to measure the quality of a split.
        """
        super().__init__(batch_size, internal_seed)

        self.n_estimators = n_estimators
        self.criterion = criterion

        if candidate_pool_size is not None:
            self.candidate_pool_size = candidate_pool_size
        else:
            self.candidate_pool_size = 1000 * batch_size

    def single_sample(  # noqa
        self,
        seed: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        raise NotImplementedError(
            "for RandomForestSampler the parallelization is hard coded in sample"
        )

    def sample(
        self,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample from the search space.

        Args:
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters
        """
        # Get large candidate pool
        candidates = self._get_candidates(existing_points, search_space)

        # Train surrogate
        x: NDArray[np.float64]
        y: NDArray[np.int64]
        x, y, _existing_points_quantiles = self.prepare_data_for_classifier(
            existing_points, existing_losses
        )

        classifier: RandomForestClassifier = RandomForestClassifier(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            n_jobs=-1,
            random_state=self.internal_seed,
        )
        classifier.fit(x, y)
        # Predict quantiles
        predicted_points_quantiles: NDArray[np.float64] = classifier.predict(candidates)
        # Sort params by predicted quantile
        sorting_indices: NDArray[np.float64] = np.argsort(predicted_points_quantiles)
        sampled_points: NDArray[np.float64] = candidates[sorting_indices][
            : self.batch_size
        ]

        self.internal_seed += self.batch_size

        return digitize_data(sampled_points, search_space.param_grid)

    @staticmethod
    def prepare_data_for_classifier(
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
        num_bins: int = 10,
    ) -> Tuple[NDArray[np.float64], NDArray[np.int64], NDArray[np.float64]]:
        """
        Prepare data for the classifier.

        Args:
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters
            num_bins: the number of bins

        Returns:
            A triple (x, y, quantiles), where
                - x is the vector of training data
                - y is the vector of targets
                - the quantiles
        """
        x: NDArray[np.float64] = existing_points
        y: NDArray[np.float64] = existing_losses

        cutoffs: NDArray[np.float64] = np.linspace(0, 1, num_bins + 1)
        quantiles: NDArray[np.float64] = np.zeros(num_bins + 1)

        for i in range(num_bins - 1):
            quantiles[i + 1] = np.quantile(y, cutoffs[i + 1])

        quantiles[-1] = np.max(y)

        y_cat: NDArray[np.int64] = np.digitize(y, quantiles, right=True)
        y_cat = y_cat - 1

        return x, y_cat, quantiles

    def _get_candidates(
        self, existing_points: NDArray[np.float64], search_space: SearchSpace
    ) -> NDArray[np.float64]:
        """
        Get the set of candidate parameters.

        Args:
            existing_points: the parameters already sampled
            search_space: the search space

        Returns:
            the list of candidates, without duplicates.
        """

        def sample(nb_samples: int) -> NDArray[np.float64]:
            random_generator = default_rng(self.internal_seed)
            candidates = np.zeros((nb_samples, search_space.dims))
            for i, params in enumerate(search_space.param_grid):
                candidates[:, i] = random_generator.choice(params, size=(nb_samples,))
            return candidates

        samples = sample(self.candidate_pool_size)

        for n in range(self.max_duplication_passes):

            duplicates = self.find_and_get_duplicates(samples, existing_points)

            num_duplicates = len(duplicates)

            if num_duplicates == 0:
                break
            new_samples = sample(num_duplicates)
            samples[duplicates] = new_samples

            if n == self.max_duplication_passes - 1:
                print(
                    f"Warning: Repeated samples still found after {self.max_duplication_passes} duplication passes."
                    " This is probably due to a small search space."
                )

        return samples



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""This module contains helper functions for plotting TS descriptive statistics."""
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import statsmodels.api as sm


def ts_stats(ts: List[float]) -> None:  # pylint: disable=too-many-locals
    """Show TS graphical descriptive statistics."""
    color = "darkslateblue"
    alpha = 0.8

    Ts = np.array(ts, dtype="double")
    SqDiff = np.append(
        np.absolute(np.diff(Ts)), 0
    )  # not precise! shouldn't append '0'!

    TsAcf = sm.tsa.acf(Ts, nlags=20)
    SqDiffAcf = sm.tsa.acf(SqDiff, nlags=20)

    TsPAcf = sm.tsa.pacf(Ts, nlags=20)
    SqDiffPAcf = sm.tsa.pacf(SqDiff, nlags=20)

    WINDOW_W = 20

    fig, _ax = plt.subplots(3, 4, figsize=(WINDOW_W, 15), sharex=False, sharey=False)

    sp1 = plt.subplot(3, 4, 1)
    plt.hist(Ts, 50, facecolor=color, alpha=alpha)
    sp1.set_title("Ts hist")

    sp2 = plt.subplot(3, 4, 2)
    plt.plot(TsAcf, marker="o", linestyle="None", c=color, alpha=alpha)
    sp2.set_title("Ts acf")

    sp3 = plt.subplot(3, 4, 3)
    plt.plot(TsPAcf, marker="o", linestyle="None", c=color, alpha=alpha)
    sp3.set_title("Ts pacf")

    sp4 = plt.subplot(3, 4, 4)
    plt.plot(Ts[0:1000], c=color, alpha=alpha)
    sp4.set_title("Ts sample")

    sp5 = plt.subplot(3, 4, 5)
    plt.hist(SqDiff, 50, facecolor=color, alpha=alpha)
    sp5.set_title("Abs 1st diff hist")

    sp6 = plt.subplot(3, 4, 6)
    plt.plot(SqDiffAcf, marker="o", linestyle="None", c=color, alpha=alpha)
    sp6.set_title("Abs 1st diff acf")

    sp7 = plt.subplot(3, 4, 7)
    plt.plot(SqDiffPAcf, marker="o", linestyle="None", c=color, alpha=alpha)
    sp7.set_title("Abs 1st diff pacf")

    sp8 = plt.subplot(3, 4, 8)
    plt.plot(SqDiff[0:1000], c=color, alpha=alpha)
    sp8.set_title("Abs 1st diff sample")

    sp9 = plt.subplot(3, 4, 9)
    for i in range(len(Ts) - 3):
        plt.plot(Ts[i + 1 : i + 3], Ts[i : i + 2], alpha=0.10, c=color)
    sp9.set_title("Ts X(t) vs X(t-1) traj")

    sp10 = plt.subplot(3, 4, 10)
    for i in range(len(Ts) - 4):
        plt.plot(SqDiff[i + 1 : i + 3], SqDiff[i : i + 2], alpha=0.10, c=color)
    sp10.set_title("Abs 1st diff X(t) vs X(t-1) traj")

    sp11 = plt.subplot(3, 4, 11)
    plt.plot(
        Ts[1 : len(Ts)],
        Ts[0 : len(Ts) - 1],
        marker="o",
        linestyle="None",
        alpha=0.10,
        c=color,
    )
    sp11.set_title("Ts X(t) vs X(t-1) params")

    sp12 = plt.subplot(3, 4, 12)
    plt.plot(
        SqDiff[1 : len(Ts)],
        SqDiff[0 : len(Ts) - 1],
        marker="o",
        linestyle="None",
        alpha=0.10,
        c=color,
    )
    sp12.set_title("Abs 1st diff X(t) vs X(t-1) params")

    fig.suptitle("Descriptive stats", fontsize=16)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation fo the Halton sampler."""

import numpy as np
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace


class HaltonSampler(BaseSampler):
    """
    Halton low discrepancy sequence.

    This snippet implements the Halton sequence following the generalization of
    a sequence of *Van der Corput* in n-dimensions.
    """

    def single_sample(
        self,
        seed: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample a single point uniformly within the search space.

        Args:
            seed: random seed
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled (not used)
            existing_losses: the loss corresponding to the sampled parameters (not used)

        Returns:
            the parameter sampled
        """
        sampled_point: NDArray[np.float64] = np.zeros(
            shape=search_space.dims, dtype=np.float64
        )
        unit_cube_points: NDArray[np.float64] = HaltonSampler.halton(
            seed=seed, dims=search_space.dims
        )

        p_bounds: NDArray[np.float64] = search_space.parameters_bounds
        for param_index in range(search_space.dims):
            sampled_point[param_index] = p_bounds[0][param_index] + unit_cube_points[
                param_index
            ] * (p_bounds[1][param_index] - p_bounds[0][param_index])

        return sampled_point

    @staticmethod
    def primes_from_2_to_n(n: int) -> NDArray[np.int64]:
        """
        Prime numbers from 2 to n.

        From `StackOverflow <https://stackoverflow.com/questions/2068372>`_.

        Args:
            n: sup bound with ``n >= 6``.

        Returns:
            primes in 2 <= p < n.
        """
        sieve = np.ones(n // 3 + (n % 6 == 2), dtype=np.bool_)
        for i in range(1, int(n**0.5) // 3 + 1):
            if sieve[i]:
                k = 3 * i + 1 | 1
                sieve[k * k // 3 :: 2 * k] = False
                sieve[k * (k - 2 * (i & 1) + 4) // 3 :: 2 * k] = False
        return np.r_[2, 3, ((3 * np.nonzero(sieve)[0][1:] + 1) | 1)]

    @staticmethod
    def van_der_corput(
        sample_size: int, base: int = 2, n_start: int = 0
    ) -> NDArray[np.float64]:
        """
        Van der Corput sequence, generalized as to accept a starting point in the sequence.

        Args:
            sample_size:  number of element of the sequence
            base: base of the sequence
            n_start: starting point of the sequence

        Returns:
            sequence of Van der Corput
        """
        sequence: NDArray[np.float64] = np.zeros(shape=sample_size - n_start)
        for index in range(n_start, sample_size):
            n_th_number: float = 0.0
            denom: float = 1.0
            i: int = index
            while i > 0:
                i, remainder = divmod(i, base)
                denom *= base
                n_th_number += remainder / denom
            sequence[index - n_start] = n_th_number
        return sequence

    @staticmethod
    def halton(seed: int = 0, dims: int = 1) -> NDArray[np.float64]:
        """
        Halton sequence.

        Changed in order to accept a starting point in the sequence and to adapt to the calibration program.

        Args:
            seed: seed of sequence
            dims: dimension

        Returns:
            sequence of Halton.
        """
        big_number: int = 10
        while "Not enough primes":
            bases: NDArray[np.int64] = HaltonSampler.primes_from_2_to_n(big_number)[
                :dims
            ]
            if len(bases) == dims:
                break
            big_number += 1000

        n_samples: int = 1
        # Generate a sample using a Van der Corput sequence per dimension.
        sample: NDArray[np.float64] = np.zeros(shape=dims)
        for i, prime in enumerate(bases):
            sample[i] = HaltonSampler.van_der_corput(
                sample_size=n_samples + 1 + seed, base=prime, n_start=seed
            )[1:]

        return sample



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the R-sequence sampler."""

import numpy as np
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace


class RSequenceSampler(BaseSampler):
    """The R-sequence sampler."""

    @staticmethod
    def r_sequence(seed: int, dims: int) -> NDArray[np.float64]:
        """
        Build R-sequence (http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/).

        Args:
            seed: seed of the sequence
            dims: Size of space.

        Returns:
            Array of params uniformly placed in d-dimensional unit cube.
        """
        phi: float = 2.0
        for _ in range(10):
            phi = pow(1 + phi, 1.0 / (dims + 1))

        # FROM KK:
        # alpha = np.array([pow(1./phi, i+1) for i in range(d)])  # flake8: noqa
        # params = np.array([(0.5 + alpha*(i+1)) % 1 for i in range(nStart, nStop)])  # flake8: noqaq

        # FROM ORIGINAL ARTICLE:
        alpha: NDArray[np.float64] = np.zeros(dims, dtype=np.float64)
        for i in range(dims):
            alpha[i] = pow(1 / phi, i + 1) % 1

        points: NDArray[np.float64] = (0.5 + alpha * (seed + 1)) % 1

        return points

    def single_sample(
        self,
        seed: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample a single point uniformly within the search space.

        Args:
            seed: random seed
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled (not used)
            existing_losses: the loss corresponding to the sampled parameters (not used)

        Returns:
            the parameter sampled
        """
        sampled_point: NDArray[np.float64] = np.zeros(
            shape=search_space.dims, dtype=np.float64
        )
        unit_cube_points: NDArray[np.float64] = RSequenceSampler.r_sequence(
            seed=seed, dims=search_space.dims
        )

        p_bounds: NDArray[np.float64] = search_space.parameters_bounds
        for param_index in range(search_space.dims):
            sampled_point[param_index] = p_bounds[0][param_index] + unit_cube_points[
                param_index
            ] * (p_bounds[1][param_index] - p_bounds[0][param_index])

        return sampled_point
